local matrix = {}
local framecache = {}
local actionqueue = {}
local functions = {}
local pop, push

local function updateMatrix()
	for i = -1, 9 do matrix[i] = wipe(matrix[i] or {}) end
	wipe(framecache)
	for i = 1, select("#", _G["BejeweledGame"]:GetChildren()) do
		local v = select(i, _G["BejeweledGame"]:GetChildren())
		if(v.gridX) then
			framecache[(v.gridY*8)+v.gridX] = v
			matrix[v.gridX][v.gridY] = v.contents
		end
	end
	push("findNextMove")
end

local function check(x, y, val)
	if not matrix[y] or not matrix[y][x] then return false end
	return matrix[y][x] == val
end

local function click(x, y)
	local f = framecache[x*8+y]
	f:GetScript("OnMouseDown")(f)
end

do
	local args = {}
	function pop()
		if #actionqueue == 0 then return end
		wipe(args)
		local func
		for i = 1, 100 do
			local val = tremove(actionqueue, 1)
			if not val then break end
			if val == "!" then break end
			if not func then				
				func = functions[val]
			else
				tinsert(args, val)
			end
		end
		if func then
			func(unpack(args))
		end
	end

	function push(...)
		for i = 1, select("#", ...) do
			local val = select(i, ...)
			tinsert(actionqueue, val)
		end
		tinsert(actionqueue, "!")
	end
end

local function move(x, y, nx, ny, strat)
	push("click", x, y)
	push("click", nx, ny)
	push("updateMatrix")
end

local function findNextMove()
	for y, row in ipairs(matrix) do
		for x, val in ipairs(row) do
			if check(x+1, y, val) then
				if 			check(x+2, y-1, val) then return move(x+2, y-1, x+2, y, 1)
				elseif 	check(x+2, y+1, val) then return move(x+2, y+1, x+2, y, 2)
				elseif 	check(x+3, y,   val) then return move(x+3, y,   x+2, y, 3)
				end
			elseif check(x-1, y, val) then
				if 			check(x-2, y-1, val) then return move(x-2, y-1, x-2, y, 4)
				elseif 	check(x-2, y+1, val) then return move(x-2, y+1, x-2, y, 5)
				elseif 	check(x-3, y,   val) then return move(x-3, y,   x-2, y, 6)
				end
			elseif check(x, y-1, val) then
				if 			check(x-1, y-2, val) then return move(x-1, y-2, x, y-2, 7)
				elseif 	check(x+1, y-2, val) then return move(x+1, y-2, x, y-2, 8)
				elseif 	check(x,   y-3, val) then return move(x,   y-3, x, y-2, 9)
				end
			elseif check(x, y+1, val) then
				if 			check(x-1, y+2, val) then return move(x-1, y+2, x, y+2, 10)
				elseif 	check(x+1, y+2, val) then return move(x+1, y+2, x, y+2, 11)
				elseif 	check(x,   y+3, val) then return move(x,   y+3, x, y+2, 12)
				end
			elseif check(x+2, y, val) then
				if			check(x+1, y-1, val) then return move(x+1, y-1, x+1, y, 13)
				elseif	check(x+1, y+1, val) then return move(x+1, y+1, x+1, y, 14)
				end
			elseif check(x-2, y, val) then
				if			check(x-1, y-1, val) then return move(x-1, y-1, x-1, y, 15)
				elseif	check(x-1, y+1, val) then return move(x-1, y+1, x-1, y, 16)
				end
			elseif check(x, y-2, val) then
				if			check(x-1, y-1, val) then return move(x-1, y-1, x, y-1, 17)
				elseif	check(x+1, y-1, val) then return move(x+1, y-1, x, y-1, 18)
				end
			elseif check(x, y+2, val) then
				if			check(x-1, y+1, val) then return move(x-1, y+1, x, y+1, 19)
				elseif	check(x+1, y+1, val) then return move(x+1, y+1, x, y+1, 20)
				end
			end
		end
	end
	push("updateMatrix")
end

functions.updateMatrix = updateMatrix
functions.click = click
functions.findNextMove = findNextMove

do
	local throttle = 0.1
	local f = CreateFrame("Frame")
	local totalElapsed = 0
	f:SetScript("OnUpdate", function(self, elapsed)
		totalElapsed = totalElapsed + elapsed
		if totalElapsed > throttle then
			totalElapsed = totalElapsed - throttle
			pop()
		end
	end)
	f:Show()
end

local solving = false
function Bejeweledize()
	wipe(actionqueue)
	push("updateMatrix")
end
SLASH_BJSOLVE1 = "/bjsolve"
SlashCmdList["BJSOLVE"] = function()
	if solving then
		print("No longer solving Bejeweled")
		wipe(actionqueue)
		solving = false
	else
		solving = true
		print("Now solving Bejeweled")
		Bejeweledize()
	end	
end
